from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_RLC(self):

        a = Circuit("""
        V 1 0 {v(t)}; down
        R1 1 2; right
        L 2 3; right=1.5, i={i_L}
        R2 3 0_3; down=1.5, i={i_{R2}}, v={v_{R2}}
        W 0 0_3; right
        W 3 3_a; right
        C 3_a 0_4; down, i={i_C}, v={v_C}
        W 0_3 0_4; right""")
        ss = a.ss

        self.assertEqual(ss.x[0], current('i_L(t)'), "x[0]")        
        self.assertEqual(ss.x[1], voltage('v_C(t)'), "x[1]")

        self.assertEqual(ss.x0[0], 0, "x0[0]")        
        self.assertEqual(ss.x0[1], 0, "x0[1]")

        self.assertEqual(ss.y[0], voltage('v_1(t)'), "y[0]")        

        self.assertEqual(ss.u[0], voltage('v(t)'), "u[0]")

        self.assertEqual(ss.A[0, 0], expr('-R1/L'), "A[0, 0]")                
        



"""This module provides the AngularFourierDomainExpression class to
represent omega-domain (angular frequency Fourier domain) expressions.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import AngularFourierDomain
from .fourier import inverse_fourier_transform
from .expr import Expr, expr
from .sym import fsym, ssym, tsym, omegasym, omega0sym, j, pi
from sympy import Expr as symExpr

__all__ = ('omegaexpr', )

class AngularFourierDomainExpression(AngularFourierDomain, Expr):
    """Fourier domain expression or symbol (angular frequency)."""

    var = omegasym

    def __init__(self, val, **assumptions):

        super(AngularFourierDomainExpression, self).__init__(val, **assumptions)

        if self.expr.find(ssym) != set():
            raise ValueError(
                'omega-domain expression %s cannot depend on s' % self.expr)
        if self.expr.find(tsym) != set():
            raise ValueError(
                'omega-domain expression %s cannot depend on t' % self.expr)

    def as_expr(self):
        return AngularFourierDomainExpression(self)
    
    def inverse_fourier(self, **assumptions):
        """Attempt inverse Fourier transform."""

        expr = self.subs(2 * pi * fsym)
        result = inverse_fourier_transform(expr, fsym, tsym)

        return self.change(result, 'time')

    def time(self, **assumptions):
        """Alias for inverse_fourier."""

        return self.inverse_fourier(**assumptions)

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""

        return self
    
    def fourier(self, **assumptions):
        """Convert to Fourier domain."""
        
        from .symbols import f
        
        result = self.subs(2 * pi * f)
        return result

    def laplace(self, **assumptions):
        """Convert to Laplace domain."""

        result = self.time(**assumptions).laplace()
        return result

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)

    def plot(self, wvector=None, **kwargs):
        """Plot angular frequency response at values specified by wvector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.
        """

        from .plot import plot_angular_frequency
        return plot_angular_frequency(self, wvector, **kwargs)

        
def omegaexpr(arg, **assumptions):
    """Create AngularFourierDomainExpression object.  If `arg` is omegasym return omega"""

    if arg is omegasym:
        return omega
    if arg is omega0sym:
        return omega0    
    return AngularFourierDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses

classes = expressionclasses.register('angular fourier', AngularFourierDomainExpression)
        
omega = AngularFourierDomainExpression('omega')
# This represents a specific angular frequency and is assumed to be positive
omega0 = AngularFourierDomainExpression('omega_0', positive=True)




"""This module provides support for the common aspects of Circuit and
Network classes.

Copyright 2014--2020 Michael Hayes, UCECE

"""

# TODO: Add option to defer evaluation and thus keep things symbolic.
# This will help to simplify results that are not cancelled due to
# numerical quantisation.

# TODO: This needs an overhaul to untangle the classes.

from __future__ import division
from .superpositionvoltage import SuperpositionVoltage
from .superpositioncurrent import SuperpositionCurrent
from .current import Iname
from .schematic import Schematic
from .netlistmixin import NetlistMixin
from .netfile import NetfileMixin
from .expr import Expr, expr
from .subnetlist import SubNetlist
from .mna import MNAMixin, Nodedict, Branchdict
from .symbols import omega
from copy import copy



class Transformdomains(dict):

    def __getattr__(self, attr):
        if attr not in self:
            raise AttributeError('Unknown attribute %s' % attr)
        return self[attr]

    def __getitem__(self, key):
        if key == 'w':
            key = omega
        # This allows a[omega] to work if omega used as key
        # instead of 'omega'.
        if isinstance(key, Expr):
            key = key.expr
        return super(Transformdomains, self).__getitem__(key)

    
class Netlist(NetlistMixin, NetfileMixin):
    """This class handles a generic netlist with multiple sources.
    During analysis, subnetlists are created for each source kind (dc,
    ac, transient, etc).  Since linearity is assumed, superposition is
    employed.

    """

    def __init__(self, filename=None, context=None, allow_anon=False):

        super (Netlist, self).__init__(filename, context, allow_anon=allow_anon)
        self._invalidate()
        self.kind = 'super'

    def _invalidate(self):

        for attr in ('_sch', '_sub', '_Vdict', '_Idict', '_analysis',
                     '_node_map', '_ss', '_node_list', '_branch_list', '_G'):
            try:
                delattr(self, attr)
            except:
                pass

    def _groups(self):
        """Return dictionary of source groups keyed by domain.

        If the netlist is for an initial value problem, all the
        sources are in a single group called 'ivp'.  Any noise sources
        are ignored.

        If the netlist can be solved in the time-domain (i.e., there
        are no reactive components), all the sources are in single
        group called 'time'.

        Otherwise, the sources are decomposed and then grouped into
        the 'dc', 's', 'n*', and omega categories.  Note, a source can
        appear in multiple groups, for example, a source with voltage
        3 + u(t) will appear in both the 'dc' and 's' groups.

        """
            
        if self.is_ivp:
            
            def namelist(elements):
                return ', '.join([elt for elt in elements])

            if self.missing_ic != {}:
                print('Warning: missing initial conditions for %s' %
                      namelist(self.missing_ic))

            groups = self.independent_source_groups()                       
            newgroups = {'ivp' : []}
            for key, sources in groups.items():
                if isinstance(key, str) and key[0] == 'n':
                    print('Warning: ignoring noise source %'
                          ' for initial value problem' % sources)
                else:
                    newgroups['ivp'] += sources
            return newgroups

        elif self.is_time_domain:
            
            groups = self.independent_source_groups()            
            newgroups = {'time' : []}
            for key, sources in groups.items():
                if isinstance(key, str) and key[0] == 'n':
                    newgroups[key] = sources
                else:
                    newgroups['time'] += sources
            return newgroups

        else:
            return self.independent_source_groups(transform=True)        

        
    def _sub_make(self):

        groups = self._groups()
        self._sub = Transformdomains()

        for kind, sources in groups.items():
            self._sub[kind] = SubNetlist(self, kind)

        return self._sub
        
    @property
    def sub(self):
        """Return dictionary of subnetlists keyed by transform domain kind.
        Note, the subnetlists are not created until a specific one is
        selected.

        """

        if hasattr(self, '_sub'):
            return self._sub

        return self._sub_make()

    @property
    def subcircuits(self):
        """Return dictionary of subnetlists keyed by transform domain kind.
        Note, the subnetlists are not created until a specific one is
        selected.  The subcircuit keys are :
        'ivp' for an initial value problem solved using Laplace methods,
        's' for transient analysis using Laplace methods,
        'dc' for DC analysis,
        'time' for time-domain analysis when there are no reactive
        components, 
        'n*' for noise-analysis (there is a subcircuit for
        each independent noise source), and 
        omega (where omega is a number of expression specifying the angular
        frequency) for phasor analysis.

        """        
        return self.sub

    @property
    def kinds(self):
        """Return list of transform domain kinds required to analyse the netlist."""
        return list(self.sub.keys())
    
    @property
    def Vdict(self):
        """Return dictionary of node voltages for each transform domain"""

        try:
            return self._Vdict
        except AttributeError:
            pass        

        result = Nodedict()
        for sub in self.sub.values():
            for node, value in sub.Vdict.items():
                if node not in result:
                    result[node] = SuperpositionVoltage()
                result[node].add(value)

        self._Vdict = result
        return result

    @property
    def Idict(self):
        """Return dictionary of branch currents for each transform domain"""

        try:
            return self._Idict
        except AttributeError:        
            pass

        result = Branchdict()
        for sub in self.sub.values():
            for node, value in sub.Idict.items():
                if node not in result:
                    result[node] = SuperpositionCurrent()
                result[node].add(value)
        self._Idict = result                    
        return result    

    def get_I(self, name):
        """Current through component"""

        result = SuperpositionCurrent()
        for sub in self.sub.values():
            I = sub.get_I(name)
            result.add(I)
        result = result.canonical()            
        return result

    def get_i(self, name):
        """Time-domain current through component"""

        return self.get_I(name).time()

    def _get_Vd(self, Np, Nm=None):
        """This does not check nodes."""
        
        result = SuperpositionVoltage()
        for sub in self.sub.values():
            Vd = sub.get_Vd(Np, Nm)
            result.add(Vd)
        result = result.canonical()
        return result

    def get_Vd(self, Np, Nm=None):
        """Voltage drop between nodes"""

        Np, Nm = self._check_nodes(Np, Nm)
        return self._get_Vd(Np, Nm)

    def get_vd(self, Np, Nm=None):
        """Time-domain voltage drop between nodes"""

        return self.get_Vd(Np, Nm).time()

    def dc(self):
        """Return subnetlist for dc components of independent sources.

        See also, ac, transient, laplace.
        """
        return SubNetlist(self, 'dc')

    def ac(self):
        """Return subnetlist for ac components of independent sources
        for angular frequency omega.

        See also, dc, transient, laplace.
        """
        # Could look at all the ac frequencies and if there is only
        # one use that?  If have multiple ac frequencies should issue
        # warning.
        return SubNetlist(self, omega)    

    def transient(self):
        """Return subnetlist for transient components of independent
        sources.  Note, unlike the similar laplace method, dc and ac 
        components are ignored.

        See also, dc, ac, laplace.

        """        
        return SubNetlist(self, 's')

    def laplace(self):
        """Return subnetlist for Laplace representations of independent
        source values.

        See also, dc, ac, transient.
        
        """        
        return SubNetlist(self, 'laplace')    
    
    



from nose.tools import *
import sys
sys.path.append('..')

from lcapy.parser import Parser
import lcapy.schemcpts as schemcpts
import lcapy.grammar as grammar

parser = Parser(schemcpts, grammar)
parse = parser.parse

@raises(ValueError)
def test_Exception1():
    '''Test missing arg'''

    parse('V1 2')

@raises(ValueError)
def test_Exception2():
    '''Test too many args'''

    parse('V1 2 3 4 5 6')

@raises(ValueError)
def test_Exception3():
    '''Test too many args'''

    parse('V1 2 3 dc 4 5 6')

@raises(ValueError)
def test_Exception4():
    '''Test unknown component'''

    parse('A1 1 2')

def test_V1():
    '''Test voltage source'''
    
    assert_equals(type(parse('V1 1 2')), schemcpts.classes['V'], 'Class not V')

def test_Vdc1():
    '''Test dc voltage source'''
    
    assert_equals(type(parse('V1 1 2 dc')), schemcpts.classes['Vdc'], 'Class not Vdc')

def test_Vac1():
    '''Test ac voltage source'''
    
    assert_equals(type(parse('V1 1 2 ac')), schemcpts.classes['Vac'], 'Class not Vac')

def test_Vsin1():
    '''Test sin voltage source'''
    
    assert_equals(type(parse('V1 1 2 sin(1, 2, 3)')), schemcpts.classes['Vsin'], 'Class not Vsin')

def test_Vexpr():
    '''Test voltage source with expr'''
    
    assert_equals(type(parse('V1 1 2 {a * 5}')), schemcpts.classes['V'], 'Class not V')

def test_Vac2():
    '''Test ac voltage source'''
    
    assert_equals(type(parse('V1 1 2 ac {1j} {0} 3')), schemcpts.classes['Vac'], 'Class not Vac')    

def test_Vquotes():
    '''Test voltage source with arg in quotes'''
    
    assert_equals(type(parse('V1 1 2 "a * 5"')), schemcpts.classes['V'], 'Class not V')

# def test_opamp():
#     '''Test opamp'''
#    
#     assert_equals(type(parse('E 1 2 opamp 3 4')), schemcpts.classes['Eopamp'], 'Class not Eopamp')



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_time_units(self):

        self.assertEqual(voltage(texpr(1)).units, 'V', 'time voltage')
        self.assertEqual(current(texpr(1)).units, 'A', 'time current')
        self.assertEqual(admittance(texpr(1)).units, 'S/s', 'time admittance')        
        self.assertEqual(impedance(texpr(1)).units, 'ohm/s', 'time impedance')
        self.assertEqual((voltage(texpr(1))**2).units, 'V^2', 'time voltage squared')
        self.assertEqual((current(texpr(1))**2).units, 'A^2', 'time current squared')
        self.assertEqual((admittance(texpr(1))**2).units, 'S^2/s^2', 'time admittance squared')                
        self.assertEqual((impedance(texpr(1))**2).units, 'ohm^2/s^2', 'time impedance squared')                

    def test_fourier_units(self):

        self.assertEqual(voltage(fexpr(1)).units, 'V/Hz', 'fourier voltage')
        self.assertEqual(current(fexpr(1)).units, 'A/Hz', 'fourier current')
        self.assertEqual(admittance(fexpr(1)).units, 'S', 'fourier admittance')        
        self.assertEqual(impedance(fexpr(1)).units, 'ohm', 'fourier impedance')
        self.assertEqual((voltage(fexpr(1))**2).units, 'V^2/Hz^2', 'fourier voltage squared')
        self.assertEqual((current(fexpr(1))**2).units, 'A^2/Hz^2', 'fourier current squared')
        self.assertEqual((admittance(fexpr(1))**2).units, 'S^2', 'fourier admittance squared')                
        self.assertEqual((impedance(fexpr(1))**2).units, 'ohm^2', 'fourier impedance squared')

    def test_constant_units(self):

        self.assertEqual(voltage(cexpr(1)).units, 'V', 'constant voltage')
        self.assertEqual(current(cexpr(1)).units, 'A', 'constant current')
        self.assertEqual(admittance(cexpr(1)).units, 'S', 'constant admittance')        
        self.assertEqual(impedance(cexpr(1)).units, 'ohm', 'constant impedance')
        self.assertEqual((voltage(cexpr(1))**2).units, 'V^2', 'constant voltage squared')
        self.assertEqual((current(cexpr(1))**2).units, 'A^2', 'constant current squared')
        self.assertEqual((admittance(cexpr(1))**2).units, 'S^2', 'constant admittance squared')                
        self.assertEqual((impedance(cexpr(1))**2).units, 'ohm^2', 'constant impedance squared')                
        
    def test_phasor_units(self):

        self.assertEqual(voltage(phasor(1)).units, 'V', 'phasor voltage')
        self.assertEqual(current(phasor(1)).units, 'A', 'phasor current')
        self.assertEqual(admittance(phasor(1)).units, 'S', 'phasor admittance')        
        self.assertEqual(impedance(phasor(1)).units, 'ohm', 'phasor impedance')
        self.assertEqual((voltage(phasor(1))**2).units, 'V^2', 'phasor voltage squared')
        self.assertEqual((current(phasor(1))**2).units, 'A^2', 'phasor current squared')
        self.assertEqual((admittance(phasor(1))**2).units, 'S^2', 'phasor admittance squared')                
        self.assertEqual((impedance(phasor(1))**2).units, 'ohm^2', 'phasor impedance squared')

    def test_discrete_time_units(self):

        self.assertEqual(voltage(nexpr(1)).units, 'V', 'nexpr voltage')
        self.assertEqual(current(nexpr(1)).units, 'A', 'nexpr current')
        self.assertEqual(admittance(nexpr(1)).units, 'S', 'nexpr admittance')        
        self.assertEqual(impedance(nexpr(1)).units, 'ohm', 'nexpr impedance')
        self.assertEqual((voltage(nexpr(1))**2).units, 'V^2', 'nexpr voltage squared')
        self.assertEqual((current(nexpr(1))**2).units, 'A^2', 'nexpr current squared')
        self.assertEqual((admittance(nexpr(1))**2).units, 'S^2', 'nexpr admittance squared')                
        self.assertEqual((impedance(nexpr(1))**2).units, 'ohm^2', 'nexpr impedance squared')

    def test_discrete_fourier_units(self):

        self.assertEqual(voltage(kexpr(1)).units, 'V', 'kexpr voltage')
        self.assertEqual(current(kexpr(1)).units, 'A', 'kexpr current')
        self.assertEqual(admittance(kexpr(1)).units, 'S', 'kexpr admittance')        
        self.assertEqual(impedance(kexpr(1)).units, 'ohm', 'kexpr impedance')
        self.assertEqual((voltage(kexpr(1))**2).units, 'V^2', 'kexpr voltage squared')
        self.assertEqual((current(kexpr(1))**2).units, 'A^2', 'kexpr current squared')
        self.assertEqual((admittance(kexpr(1))**2).units, 'S^2', 'kexpr admittance squared')                
        self.assertEqual((impedance(kexpr(1))**2).units, 'ohm^2', 'kexpr impedance squared')                        



from lcapy import *
from lcapy.discretetime import *
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_ztransform(self):

        self.assertEqual(nexpr(1).ZT(), z / (z - 1), "1")
        self.assertEqual(nexpr(3).ZT(), 3 * z / (z - 1), "3")
        self.assertEqual(Heaviside(n).ZT(), z / (z - 1), "Heaviside(n)")
        self.assertEqual(3 * Heaviside(n).ZT(), 3 * z / (z - 1), "3 * Heaviside(n)")
        self.assertEqual(Heaviside(n + 1).ZT(), z **2 / (z - 1), "Heaviside(n + 1)")
        self.assertEqual(Heaviside(n - 1).ZT(), 1 / (z - 1), "Heaviside(n - 1)")        
        self.assertEqual(unitimpulse(n).ZT(), 1, "unitimpulse(n)")
        self.assertEqual(3 * unitimpulse(n).ZT(), 3, "unitimpulse(n)")        
        self.assertEqual(unitimpulse(n - 1).ZT(), 1 / z, "unitimpulse(n - 1)")
        self.assertEqual(unitimpulse(n - 2).ZT(), 1 / z**2, "unitimpulse(n - 2)")
        self.assertEqual(3 * unitimpulse(n - 2).ZT(), 3 / z**2, "3 * unitimpulse(n - 2)")
        self.assertEqual(3 * unitimpulse(n + 2).ZT(), 3 * z**2, "3 * unitimpulse(n + 2)")                
        self.assertEqual(expr('v(n)').ZT(), expr('V(z)'), "v(n)")
        self.assertEqual(expr('v(n / 3)').ZT(), expr('V(z**3)'), "v(n/3)")
        self.assertEqual(expr('3 * v(n / 3)').ZT(), expr('3 * V(z**3)'), "3 * v(n/3)")
        self.assertEqual(expr('v(n-3)').ZT(), expr('V(z) / z**3'), "v(n - 3)")
        self.assertEqual(nexpr('x(n)').ZT(), zexpr('X(z)'), "x(n)")

        self.assertEqual((0.1**n).ZT(), 10 * z / (10 * z - 1), "0.1**n")
        self.assertEqual(((0.1**n) * u(n)).ZT(), 10 * z / (10 * z - 1), "0.1**n")        
                         

    def test_inverse_ztransform(self):

        self.assertEqual(zexpr(1).IZT(causal=True), unitimpulse(n), "1")
        self.assertEqual((z**-1).IZT(causal=True), unitimpulse(n-1), "z**-1")
        self.assertEqual((z**-2).IZT(causal=True), unitimpulse(n-2), "z**-2")
        self.assertEqual(zexpr('1 / (1 - a * z ** -1)').IZT(causal=True), nexpr('a**n * u(n)'), "1 / (1 - a * z)")                        
        self.assertEqual(zexpr('X(z)').IZT(causal=True), nexpr('x(n)'), "X(z)")


    def test_misc(self):

        x = 3 * unitimpulse(n - 2)
        y1 = x.differentiate()
        y2 = x.ZT().ndifferentiate().IZT()
        self.assertEqual(y1, y2, "differentiate")

        x = 3 * unitimpulse(n - 2)
        y1 = x.integrate()
        y2 = x.ZT().nintegrate().IZT()
        # Need to teach that these are equal
        #self.assertEqual(y1, y2, "integrate")                
        
    def test_convolution(self):

        a = expr('a(n)')
        b = expr('b(n)')        
        c = (a.ZT() * b.ZT()).IZT(causal=True)
        d = expr('Sum(a(-m + n)*b(m), (m, 0, n))')
        
        self.assertEqual(c, d, "convolution")



"""This module provides the Latex class for formatting
values and labels.

Copyright 2020 Michael Hayes, UCECE

"""


import re
from .config import subscripts

sub_super_pattern = re.compile(r"([_\^]){([a-zA-Z]+)([0-9]*)}")


def latex_mathrm(s):
        
    def foo(match):
        
        word = match.group(2)
        suffix = word + match.group(3)
        
        # Perhaps look up dictionary to find valid words?
        # Assume that if length 3 or more then a word.
        if word.lower() in subscripts or len(word) > 2:
            suffix = r'{\mathrm{%s}}' % suffix
        else:
            suffix = r'{%s}' % suffix
        return match.group(1) + suffix

    return sub_super_pattern.sub(foo, s)


class Latex(object):

    def __init__(self, string):

        self.str = string

    def mathrm(self):
        """Place words in sub- or super-scripts inside a mathrm.
        For example V_{rms} -> V_{\mathrm{rms}}"""

        return latex_mathrm(self.str)

    def __str__(self):

        s = self.str
        return latex_mathrm(s)
            

def latex_str(string):

    return Latex(string).__str__()


def latex_format_label(s):

    if s == '':
        return s

    # With leading $ and no trailing $, e.g., v=$V1, try to
    # automagically convert to LaTeX string, otherwise pass
    # verbatim.  Currently, this only converts words in sub- or
    # super- scripts to roman.  TODO, handle more cases.
    if s[0] == '$' and s[-1] != '$':
        return '$' + latex_str(s[1:]) + '$'

    # If find a $ assume that the label is correctly formatted.
    if '$' in s:
        return s

    # If have _, ^, \frac, etc.  need to be in math-mode.
    # Should prevent lcapy generating such strings and 
    # warn user to explicity use math mode.  The tricky part is that
    # arbitrary signals may have math symbols, say sqrt.
    math_symbols = ('_',  '^', '\\left', '\\math', '\\frac', '\\sqrt')
    for math_symbol in math_symbols:
        if math_symbol in s:
            return '$' + latex_str(s) + '$'
    return s


def latex_format_node_label(s):

    if s[0] == '_':
        return s.replace('_', r'\_')

    parts = s.split('_')
    if len(parts) > 2:
        raise ValueError('Double subscript in %s' % s)
    elif len(parts) == 2:
        return '%s$_{\mathrm{%s}}$' % (parts[0], parts[1])

    return s

    



"""This module performs parsing of SPICE-like netlists.  It uses a
custom parser rather than lex/yacc to give better error messages.

Copyright 2015--2020 Michael Hayes, UCECE

"""

import re

# Could use a script to generate parser and parsing tables if speed
# was important.

def split(s, delimiters):
    """Split string by specified delimiters but not if a delimiter is
    within curly brackets {} or ""."""

    parts = []
    current = []
    close_bracket = ''
    bracket_stack = []
    for c in (s + delimiters[0]):
        if c in delimiters and len(bracket_stack) == 0:
            if len(current) > 0:
                parts.append(''.join(current))
            current = []
        else:
            if c == close_bracket:
                close_bracket = bracket_stack.pop()                
            elif c == '{':
                bracket_stack.append(close_bracket)
                close_bracket = '}'
            elif c == '"':
                bracket_stack.append(close_bracket)
                close_bracket = '"'
            current.append(c)
    if close_bracket != '':
        raise ValueError('Missing %s in %s' % (close_bracket, s))
    return parts


class Param(object):

    def __init__(self, name, base, comment):
        
        self.name = name
        self.base = base
        self.comment = comment
        self.baseclass = None

    def is_valid(self, string):

        if self.baseclass is None:
            return True
        return self.baseclass.is_valid(string)


class Rule(object):

    def __init__(self, cpt_type, classname, params, comment, pos):
        
        self.type = cpt_type
        self.classname = classname
        self.params = params
        self.comment = comment
        self.pos = pos

    def __repr__(self):

        return self.type + 'name ' + ' '.join(self.params)

    def syntax_error(self, error, string):

        raise ValueError('Syntax error: %s when parsing %s\nExpected format: %s' % (error, string, repr(self)))        

    def process(self, paramdict, string, fields, name, namespace):

        params = self.params
        if len(fields) > len(params):
            extra = ''
            if '(' in string:
                extra = ' (perhaps enclose expressions with parentheses in {})'
            self.syntax_error('Too many args' + extra, string)

        nodes = []
        args = []
        for m, param in enumerate(params):

            if m >= len(fields):
                # Optional argument
                if param[0] == '[':
                    break
                self.syntax_error('Missing arg %s' % param, string)

            if param[0] == '[':
                param = param[1:-1]

            field = fields[m]

            if paramdict[param].base in ('pin', 'node'):
                if field[0] == '.':
                    # Note, name contains namespace
                    field = name + field
                else:
                    field = namespace + field
                nodes.append(field)
            elif paramdict[param].base != 'keyword':
                args.append(field)

        return tuple(nodes), args


class Parser(object):

    def __init__(self, cpts, grammar, allow_anon=False):
        """cpts is a module containing a class for each component
        grammar is a module defining the syntax of a netlist"""

        # A string defining the syntax for a netlist
        rules = grammar.rules
        # A string defining parameters
        params = grammar.params
        # A string defining delimiter characters
        self.delimiters = grammar.delimiters
        # A string defining comment characters
        self.comments = grammar.comments
        self.allow_anon = allow_anon

        self.cpts = cpts
        self.paramdict = {}
        self.ruledict = {}
        
        for param in params.split('\n'):
            self._add_param(param)

        for rule in rules.split('\n'):
            self._add_rule(rule)

        cpts = sorted(self.ruledict.keys(), key=len, reverse=True)

        # The symbol name must be a valid Sympy symbol name so
        # it cannot include symbols such as + and -.
        self.cpt_pattern = re.compile("(%s)([#_\w'?]+)?" % '|'.join(cpts))

    def _add_param(self, string):

        if string == '':
            return

        fields = string.split(':')
        paramname = fields[0]
        fields = fields[1].split(';', 1)
        parambase = fields[0].strip()
        comment = fields[1].strip()
        
        self.paramdict[paramname] = Param(paramname, parambase, comment)

    def _add_rule(self, string):

        if string == '':
            return

        fields = string.split(':')
        cpt_classname = fields[0]
        fields = fields[1].split(';', 1)
        string = fields[0].strip()
        comment = fields[1].strip()
        
        fields = string.split(' ')
        params = fields[1:]

        # Skip the name part in the rule, e.g., only consider D from Dname.
        cpt_type = fields[0][0:-4]

        pos = None
        for m, param in enumerate(params):
            if param[0] == '[':
                param = param[1:-1]
            if param not in self.paramdict:
                raise ValueError('Unknown parameter %s for %s' % (param, string))
            if pos is None and self.paramdict[param].base == 'keyword':
                pos = m

        if cpt_type not in self.ruledict:
            self.ruledict[cpt_type] = ()
        self.ruledict[cpt_type] += (Rule(cpt_type, cpt_classname,
                                        params, comment, pos), )

    def parse(self, string, namespace='', parent=None):
        """Parse string and create object"""

        directive = False
        net = string.strip()
        if net == '':
            directive = True
        elif net[0] in self.comments:
            directive = True            
        elif net[0] == ';':
            directive = True                        
        elif net[0] == '.':
            directive = True

        if directive:
            cpt_type = 'XX'
            cpt_id = ''
            name = 'XX'
            name += parent._make_anon(cpt_type)
            defname = namespace + cpt_type + cpt_id

            if string.startswith(';') and not string.startswith(';;'):
                opts_string = string[1:]
            else:
                opts_string = ''                
                
            return self.cpts.make('XX', parent, '', defname, name,
                                  cpt_type, cpt_id, string, opts_string, (), '')

        net = namespace + net
        parts = net.split(';', 1)
        
        fields = split(parts[0], self.delimiters)

        # Strip {} and "".
        for m, field in enumerate(fields):
            if field[0] in '{"':
                fields[m] = fields[m][1:-1]
        
        name = fields.pop(0)
        parts = name.split('.')
        namespace = ''
        if len(parts) > 1:
            namespace = '.'.join(parts[0:-1]) + '.'
        name = parts[-1]

        match = self.cpt_pattern.match(name)
        if match is None:
            raise ValueError('Unknown component %s while parsing "%s"' % (name, net))

        groups = match.groups()
        cpt_type, cpt_id = groups[0], groups[1]
        if cpt_id is None:
            cpt_id = ''

        # This is the most hackery aspect of this parser where we
        # choose the rule pattern based on a keyword.  If the
        # keyword is not present, default to first rule pattern.
        # Perhaps a factory should sort this out?
        rule = self.ruledict[cpt_type][0]
        keyword = ''
        for rule1 in self.ruledict[cpt_type]:
            pos = rule1.pos
            if pos is None:
                continue
            if len(fields) > pos and fields[pos].lower() == rule1.params[pos]:
                rule = rule1
                keyword = rule1.params[pos]
                break

        defname = namespace + cpt_type + cpt_id
        name = defname
        if (cpt_id == '' and parent is not None
            and (cpt_type in ('A', 'W', 'O', 'P')) or self.allow_anon):
            name += parent._make_anon(cpt_type)
        elif cpt_id == '?':
            # Automatically name cpts to ensure they are unique
            name = name[:-1] + parent._make_anon(cpt_type)

        nodes, args = rule.process(self.paramdict, net, fields, name, 
                                   namespace)

        parts = net.split(';', 1)
        opts_string = parts[1].strip() if len(parts) > 1 else '' 

        keyword = (pos, keyword)

        # self.cpts is either the mnacpts or schematic module
        return self.cpts.make(rule.classname, parent, namespace,
                              defname, name, cpt_type, cpt_id, net,
                              opts_string, tuple(nodes), keyword,
                              *args)



from lcapy import Circuit, f

a = Circuit(""" 
R 1 0; down 
W 1 2; right 
C 2 0_2; down 
W 0 0_2; right""")   

b = a.noisy()

Vn = b.C.V.n(f).subs({'R':10e3, 'C':1e-9, 'T':293, 'k_b':1.38e-23})   

from numpy import linspace
vf = linspace(0, 100e3, 200)

from matplotlib.pyplot import savefig
(Vn * 1e9).plot(vf, plot_type='mag', ylabel='ASD (nV/rootHz)')

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')




"""This module provides the FourierNoiseDomainExpression class to represent
f-domain (Fourier domain) noise expressions.

Copyright 2014--2020 Michael Hayes, UCECE

"""
from __future__ import division
from .sym import symsimplify
from .functions import sqrt
from .sym import pi, omegasym, fsym
from .state import state
from .domains import FourierNoiseDomain
from .expr import expr
from .noiseexpr import NoiseExpression
from .omegaexpr import omega, AngularFourierDomainExpression
from .fexpr import FourierDomainExpression
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
import sympy as sym
import numpy as np

class FourierNoiseDomainExpression(FourierNoiseDomain, NoiseExpression):
    """Frequency domain (one-sided) noise spectrum expression (amplitude
    spectral density).

    This characterises a wide-sense stationary, zero-mean Gaussian
    noise random process.

    When performing arithmetic on two FourierNoiseDomainExpression
    expressions it is assumed that they are uncorrelated unless they
    have the same nid (noise indentifier).  If the nid is not
    specified, a new one is created.

    Uncorrelated noise expressions are added in quadrature (on a power
    basis).  Thus (FourierNoiseDomainExpression(3) +
    FourierNoiseDomainExpression(4)).expr = 5 since 5 = sqrt(3**2 +
    4**2)

    FourierNoiseDomainExpression(3) != FourierNoiseDomainExpression(3)
    since they are different noise realisations albeit with the same
    properties.  However, FourierNoiseDomainExpression(3).expr ==
    FourierNoiseDomainExpression(3).expr.  Similarly,
    FourierNoiseDomainExpression(3, nid='n1') ==
    FourierNoiseDomainExpression(3, nid='n1') since they have the same
    noise identifier and thus have the same realisation.

    Caution: The sum of two noise expressions generates a noise
    expression with a new nid.  This can lead to unexpected results
    since noise expressions with different nids are assumed to be
    uncorrelated.  For example, consider:
    a = FourierNoiseDomainExpression(3);
    b = FourierNoiseDomainExpression(4)
    a + b - b gives sqrt(41) but a + b - a gives sqrt(34).

    This case is correctly handled by the SuperpositionVoltage and
    SuperpositionCurrent classes since each noise component is stored
    and considered separately.

    (SuperpositionVoltage(a) + SuperpositionVoltage(b) -
    SuperpositionVoltage(b)).n gives 3 as expected.

    """
    var = fsym

    def plot(self, fvector=None, **kwargs):
        """Plot frequency response at values specified by fvector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.
        """
        from .plot import plot_frequency

        return plot_frequency(self, fvector, **kwargs)

    def transform(self, arg, **assumptions):
        """Transform into a different domain."""

        arg = expr(arg)        
        if isinstance(arg, AngularFourierDomainExpression):
            result = self.subs(omega / (2 * pi))
            cls = self._class_by_quantity(self.quantity, 'angular fourier noise')
            return cls(result, nid=self.nid, **assumptions)
        elif isinstance(arg, FourierDomainExpression):
            result = self.subs(arg, **assumptions)
            cls = self._class_by_quantity(self.quantity, 'fourier noise')
            return cls(result, nid=self.nid, **assumptions)            

        return super(FourierNoiseDomainExpression, self).transform(arg, **assumptions)
    
    
class FourierNoiseDomainVoltage(VoltageMixin, FourierNoiseDomainExpression):
    """Voltage noise amplitude spectral density (units V/rtHz).
    This can be a function of linear frequency, f.  For example,
    to model an opamp voltage noise:

    v = FourierNoiseDomainVoltage(1e-8 / sqrt(f) + 8e-9)
    
    """

    quantity_label = 'Voltage noise spectral density'
    units = 'V/rtHz'


class FourierNoiseDomainCurrent(CurrentMixin, FourierNoiseDomainExpression):
    """Current noise amplitude spectral density (units A/rtHz).

    This can be a function of linear frequency, f.  For example,
    to model an opamp current noise:

    i = FourierNoiseDomainCurrent(3e-12 / sqrt(f) + 200e-15)
    """

    quantity_label = 'Current noise spectral density'
    units = 'A/rtHz'


from .expressionclasses import expressionclasses

expressionclasses.register('fourier noise', FourierNoiseDomainExpression, None,
                           ('voltage', 'current'))


from .fexpr import f
from .noiseomegaexpr import AngularFourierNoiseDomainExpression



"""This module contains the RandomNetwork class for creating random
networks.

Copyright 2020--2021 Michael Hayes, UCECE"""

from .oneport import Vdc, Idc, Vstep, Istep, Vac, Iac, R, L, C
import random


class RandomNetworkMaker(object):

    def __init__(self, num_resistors=3, num_inductors=0, num_capacitors=0,
                 num_voltage_sources=1, num_current_sources=0,
                 num_parallel=None, numeric_values=False, kind='transient'):
        """`kind` can be 'transient', 'ac', or 'dc'."""
        
        self.num_resistors = num_resistors
        self.num_inductors = num_inductors
        self.num_capacitors = num_capacitors
        self.num_voltage_sources = num_voltage_sources
        self.num_current_sources = num_current_sources
        self.num_parallel = num_parallel
        self.numeric_values = numeric_values
        self.kind = kind

        cpts = []
        cpts.extend(self._add_cpts(R, self.num_resistors))
        cpts.extend(self._add_cpts(L, self.num_inductors))
        cpts.extend(self._add_cpts(C, self.num_capacitors))

        if self.kind == 'transient':
            cpts.extend(self._add_cpts(Vstep, self.num_voltage_sources))
            cpts.extend(self._add_cpts(Istep, self.num_current_sources))
        elif self.kind == 'dc':
            cpts.extend(self._add_cpts(Vdc, self.num_voltage_sources))
            cpts.extend(self._add_cpts(Idc, self.num_current_sources))
        elif self.kind == 'ac':
            cpts.extend(self._add_cpts(Vac, self.num_voltage_sources))
            cpts.extend(self._add_cpts(Iac, self.num_current_sources))            
        else:
            raise ValueError('Unknown circuit kind %s' % self.kind)

        self.cpts = cpts
        self.num_parallel = num_parallel

        self.connections = self._choose_connections()

    def _choose_connections(self):
        Nconnections = len(self.cpts) - 1
        
        if self.num_parallel is None:
            connections = random.choices((True, False), k=Nconnections)
        else:
            num_parallel = min(self.num_parallel, Nconnections)
            Nseries = Nconnections - num_parallel
            connections = [True] * num_parallel + [False] * Nseries

        return connections

    def __call__(self):

        return self.make()

    def _make1(self):

        cpts = random.sample(self.cpts, len(self.cpts))
        connections = random.sample(self.connections, len(self.connections))

        net = cpts[0]
        for cpt, connection in zip(cpts[1:], connections):
            if connection:
                if cpt.is_voltage_source and net.has_parallel_V:
                    return None
                net = net.parallel(cpt)
            else:
                if cpt.is_current_source and net.has_series_I:
                    return None
                net = net.series(cpt)            
        return net

    def make(self):

        while True:
            net = self._make1()
            if net is not None:
                return net
            self.connections = self._choose_connections()
            
    def _add_cpts(self, cpt, num):

        cpts = []
        for m in range(num):
            if self.numeric_values:
                value = str(random.randint(1, 10))
            else:
                value = '%s%d' % (cpt.__name__[0:1], m + 1)
            
            cpts.append(cpt(value))
        return cpts

    
def random_network(num_resistors=3, num_inductors=0, num_capacitors=0,
                   num_voltage_sources=1, num_current_sources=0,
                   num_parallel=None, numeric_values=False, kind='transient'):

    return RandomNetworkMaker(num_resistors=num_resistors,
                              num_inductors=num_inductors,
                              num_capacitors=num_capacitors,
                              num_voltage_sources=num_voltage_sources,
                              num_current_sources=num_current_sources,
                              num_parallel=num_parallel,
                              numeric_values=numeric_values,
                              kind=kind).make()



"""This module provides transfer function support.

Copyright 2020 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu

def transfer(arg, **assumptions):

    expr1 = expr(arg, **assumptions)
    if expr1.is_admittance:
        return expr1.apply_unit(1 / uu.ohms)

    try:
        expr1 = expr1.as_transfer()
    except:
        raise ValueError('Cannot represent %s(%s) as transfer function' % (expr1.__class__.__name__, expr1))

    return expr1




"""This modules provides the DiscreteExpression class to provide common methods for
the discrete-time representations.

Copyright 2020 Michael Hayes, UCECE

"""

import numpy as np
from .expr import Expr

class DiscreteExpression(Expr):

    """Superclass of discrete-time, discrete-frequency, and z-domain
    expressions."""


    def __call__(self, arg, **assumptions):
        """Transform domain or substitute arg for variable. 
        
        Substitution is performed if arg is a tuple, list, numpy
        array, or constant.  If arg is a tuple or list return a list.
        If arg is an numpy array, return numpy array.

        Domain transformation is performed if arg is a domain variable
        or an expression of a domain variable.

        See also evaluate.

        """

        from .fexpr import f
        from .nexpr import n
        from .kexpr import k
        from .zexpr import z

        if isinstance(arg, (tuple, list)):
            return [self._subs1(self.var, arg1) for arg1 in arg]

        if isinstance(arg, np.ndarray):
            return np.array([self._subs1(self.var, arg1) for arg1 in arg])

        if id(arg) in (id(n), id(z), id(k), id(f)):
            return self.transform(arg, **assumptions)

        if arg in (n, k, z, f):
            return self.transform(arg, **assumptions)    

        # Do we really want to this?   
        return super(DiscreteExpression, self).__call__(arg, **assumptions)

    def transform(self, arg, **assumptions):

        from .nexpr import DiscreteTimeDomainExpression, n
        from .kexpr import DiscreteFourierDomainExpression, k
        from .zexpr import ZDomainExpression, z
        from .fexpr import f        

        # Is this wise?   It makes sense for Voltage and Impedance objects
        # but may cause too much confusion for other expressions
        if arg is n and isinstance(self, ZDomainExpression):
            return self.IZT(**assumptions)
        elif arg is n and isinstance(self, DiscreteFourierDomainExpression):
            return self.IDFT(**assumptions)        
        elif arg is z and isinstance(self, DiscreteTimeDomainExpression):
            return self.ZT(**assumptions)
        elif arg is z and isinstance(self, DiscreteFourierDomainExpression):
            return self.IDFT(**assumptions).ZT(**assumptions)
        elif arg is k and isinstance(self, DiscreteTimeDomainExpression):
            return self.DFT(**assumptions)
        elif arg is k and isinstance(self, ZDomainExpression):
            return self.IZT(**assumptions).DFT(**assumptions)
        elif arg is f and isinstance(self, DiscreteTimeDomainExpression):
            return self.DTFT(**assumptions)
        elif arg is f and isinstance(self, ZDomainExpression):
            return self.DTFT(**assumptions)                
        
        # Do we really want to this?   
        super(DiscreteExpression, self).transform(arg, **assumptions)            



"""This module provides current support.

Copyright 2020 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu
from .classmap import domain_kind_to_symbol, domain_kind_quantity_to_class

def Iname(name, kind, cache=False):

    # Not caching is a hack to avoid conflicts of Vn1 with Vn1(s) etc.
    # when using subnetlists.  The alternative is a proper context
    # switch.  This would require every method to set the context.

    if kind == 't':
        name = name.lower()
    
    undef = domain_kind_to_symbol(kind, name)
    cls = domain_kind_quantity_to_class(kind, 'current')
    return cls(undef, cache=cache)


def Itype(kind):

    return domain_kind_quantity_to_class(kind, 'current')        


def current(arg, **assumptions):

    expr1 = expr(arg, **assumptions)

    if 'nid' in assumptions:
        from .noisefexpr import FourierNoiseDomainCurrent
        from .noiseomegaexpr import AngularFourierNoiseDomainCurrent        
        
        if expr1.is_fourier_domain or expr1.is_constant_domain:
            expr1 = FourierNoiseDomainCurrent(expr1)
        elif expr1.is_angular_fourier_domain:
            expr1 = AngularFourierNoiseDomainCurrent(expr1)
        else:
            raise ValueError('Cannot represent noise current in %s domain' % expr1.domain)
    
    try:
        expr1 = expr1.as_current()
    except:        
        raise ValueError('Cannot represent %s(%s) as current' % (expr1.__class__.__name__, expr1))

    return expr1.apply_unit(uu.amperes)


def noisecurrent(arg, **assumptions):
    """Create a new noise current with specified amplitude spectral density."""    

    nid = assumptions.get('nid', None)
    positive = assumptions.get('positive', True)        

    return current(arg, nid=None, positive=positive, **assumptions)


def phasorcurrent(arg, omega=None, **assumptions):
    from .phasor import phasor
    
    return phasor(arg, omega, **assumptions).as_current()




"""This module provides the NetlistHelper class used by the
NetlistMaker and LadderMaker classes.

Copyright 2020 Michael Hayes, UCECE

"""

class NetlistHelper(object):

    evalf = False
    
    @property 
    def _node(self):

        if not hasattr(self, '_node_counter'):
            self._node_counter = 0
        ret = self._node_counter
        self._node_counter += 1
        return ret

    def _make_id(self, kind):
        """Make identifier"""

        # TODO, use ComponentNamer.  Perhaps the names
        # should try to follow the value names, say when generated
        # by the random network generator?  This will avoid confusion.
        
        if not hasattr(self, '_anon'):
            self._anon = {}            

        if kind not in self._anon:
            self._anon[kind] = 0
        self._anon[kind] += 1
        return self._anon[kind]
    
    def _netargs(self, net):

        def process(arg):

            if self.evalf:
                arg = arg.evalf(n=self.evalf)

            arg = str(arg)

            # TODO: make more robust to catch expressions.
            if ('(' in arg) or (')' in arg) or (' ' in arg) or (',' in arg) or ('*' in arg) or ('/' in arg):
                return '{%s}' % arg
            return arg

        return ' '.join([process(arg) for arg in net.args])




"""This module provides the PhasorDomain classes to represent phasors
 for AC analysis.

A phasor represents the amplitude and phase for a single sinusoid.  By
default the angular frequency is omega_0 but it can be any number or
symbol.

Phasors straddle the time and frequency domains and this is my excuse
for the confusing phasor classes.

A phasor is described by an amplitude and phase.  There is also an
implicit frequency.  The amplitude and phase are usually functions of
frequency.  A phasor is useful to describe an AC voltage or current
signal.

A ratio of two phasors (of the same frequency) is no longer a
time-domain signal but a frequency domain quantity.  The frequency
dependence is explicit.  A phasor ratio is useful to describe an
impedance, immitance, or transfer function.  These are frequency
domain concepts.  A phasor ratio can be inferred from the Laplace
domain by substituting jomega for s, where omega is the angular
frequency of the phasor.


Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .acdc import ACChecker
from .domains import PhasorTimeDomain, PhasorFrequencyDomain
from .sym import j, omegasym
from .expr import expr
from .functions import sin, cos, exp, sqrt
from .expr import Expr
from .omegaexpr import omega
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin

__all__ = ('phasor', )

# The phasor domain is different from the Fourier and Laplace domain
# since there is an implicit anglular frequency.  This is only needed
# for voltage and current (vorrent) expressions.

# The phasor domain immittance can be found from the Laplace domain
# impedance/admittance by substituting s with j * omega.  In many
# cases the result is the same as the Fourier domain immittance but
# not always.  The most important case is the impedance of a
# capacitor.  In the Laplace domain it is 1/(s C); in the phasor
# domain it is 1/(j omega C); in the Fourier domain it is 1/(2 j omega
# C) + delta(omega) / (2 C).



class PhasorExpression(Expr):

    def __init__(self, val, **assumptions):

        if isinstance(val, PhasorExpression):
            assumptions['omega'] = val.omega
        elif 'omega' not in assumptions or assumptions['omega'] is None:
            assumptions['omega'] = omegasym
        
        assumptions['ac'] = True
        super (PhasorExpression, self).__init__(val, **assumptions)
    
    @property
    def omega(self):
        """Return angular frequency."""

        return self.assumptions.get('omega', None)

    @property
    def var(self):
        """Return angular frequency."""

        return self.omega

    def fourier(self, **assumptions):
        """Fourier transform."""

        return self.time().fourier()

    def angular_fourier(self, **assumptions):
        """Angular Fourier transform."""

        if self.has(omega):
            print('Warning: expression contains omega, should substitute with a different symbol')
        
        return self.time().angular_fourier()            
    
    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        return self.time().laplace()

    def phasor(self):
        """Convert to phasor representation."""
        
        return self.__class__(self, **self.assumptions)

    def rms(self):
        """Return root mean square."""
        
        return abs(self) * sqrt(2) / 2

    def plot(self, wvector=None, **kwargs):
        """Plot polar diagram for a time-domain phasor or frequency response
        for frequency-domain phasor.  For the latter, wvector
        specifies the angular frequencies.  If it is a tuple, it sets
        the angular frequency limits."""
        
        from .plot import plot_phasor, plot_angular_frequency

        if self.is_phasor_frequency_domain:
            return plot_angular_frequency(self, wvector, **kwargs)            
        
        return plot_phasor(self, **kwargs)

    def _mul_compatible(self, x):

        if not hasattr(x, 'omega'):
            return True

        if self.omega == x.omega:
            return True
        
        raise ValueError('Incompatible phasor angular frequencies %s and %s' %
                         (self.omega, x.omega))
        

class PhasorTimeDomainExpression(PhasorTimeDomain, PhasorExpression):
    """This is a phasor domain base class for voltages and currents."""

    is_phasor_domain = True
    is_phasor_time_domain = True

    def as_expr(self):
        return PhasorTimeDomainExpression(self)

    @classmethod
    def from_time(cls, expr, omega=None, **assumptions):

        from .symbols import t

        if expr.is_admittance or expr.is_impedance or expr.is_transfer:
            print('Should convert %s expression to Laplace-domain first' % expr.quantity)

        assumptions['ac'] = True

        if expr.is_transform_domain:
            print('Warning, converting %s-domain to time-domain first.' %
                  expr.domain)
            expr = expr.time()
            
        check = ACChecker(expr, t)
        if not check.is_ac:
            raise ValueError(
                'Do not know how to convert %s to phasor.  Expecting an AC signal.' % expr)
        
        if omega is not None and check.omega != omega.expr:
            raise ValueError('Expecting omega=%s, found omega=%s.' % (omega, check.omega))

        if check.omega == 0:
            print('Warning, DC phasor.')
        
        result = check.amp * exp(j * check.phase)
        assumptions['omega'] = check.omega

        return cls.change(expr, result, domain='phasor', **assumptions)

    def time(self, **assumptions):
        """Convert to time domain representation."""
        from .symbols import t
        
        omega1 = self.omega
        if isinstance(omega1, Expr):
            # TODO: Fix inconsistency.  Sometimes omega is a symbol.
            omega1 = omega1.expr

        if self.is_complex:
            result = self.expr * exp(j * omega1 * t)
        else:
            result = self.real.expr * cos(omega1 * t) - self.imag.expr * sin(omega1 * t)

        return TimeDomainExpression(result)
    

class PhasorFrequencyDomainExpression(PhasorFrequencyDomain, PhasorExpression):
    """This represents the ratio of two-phasors; for example
    an impedance, an admittance, or a transfer function."""

    is_phasor_frequency_domain = True    
    is_transform_domain = True

    @classmethod
    def from_laplace(cls, expr, omega=None, **assumptions):

        from .symbols import s        

        if omega is None:
            omega = omegasym

        if expr.is_voltage or expr.is_current:
            print('Should convert %s expression to time-domain first' % expr.quantity)

        result = expr.laplace(**assumptions).replace(s, j * omega)
        return cls.change(expr, PhasorFrequencyDomainExpression(result, omega=omega,
                                                                **assumptions))
    
    def time(self, **assumptions):
        """Convert to time domain representation."""
        from .symbols import s
        from .sexpr import LaplaceDomainExpression

        omega = self.omega
        result = LaplaceDomainExpression(self.replace(j * omega, s))
        return self.change(result.time())

    def as_expr(self):
        return PhasorFrequencyDomainExpression(self)
    

def phasor(arg, omega=None, **assumptions):
    """Create phasor.   

    If arg has the form A * cos(w * t + phi) the phasor 
    A * exp(j * phi) of angular frequency w is returned.

    If arg is a constant C, the phasor C is created with omega as the
    angular frequency (this defaults to 'omega' if not specified).

    """

    arg = expr(arg)

    if arg.is_time_domain:
        # Expecting AC signal.
        return PhasorTimeDomainExpression.from_time(arg, omega=omega, **assumptions)
    elif arg.is_unchanging:
        # Expecting phasor (complex amplitude)        
        return PhasorTimeDomainExpression(arg, omega=omega, **assumptions)
    else:
        # Is this sensible?
        return PhasorFrequencyDomainExpression(arg, omega=omega, **assumptions)        
    

from .expressionclasses import expressionclasses

expressionclasses.register('phasor', PhasorTimeDomainExpression,
                           PhasorFrequencyDomainExpression,
                           ('voltage', 'current', 'voltagesquared', 'currentsquared'))
from .texpr import TimeDomainExpression
from .expr import Expr




"""This modules provides the SequenceExpression class to provide common methods for
the discrete-time and discrete-frequency expressions.

Copyright 2020 Michael Hayes, UCECE

"""

from .dexpr import DiscreteExpression
from .sequence import Sequence
from .functions import Heaviside, UnitStep, DiracDelta, UnitImpulse
from numpy import arange


class SequenceExpression(DiscreteExpression):
    """Superclass of discrete-time and discrete-frequency expressions."""

    def __init__(self, val, **assumptions):

        super(SequenceExpression, self).__init__(val, **assumptions)

        if self.has(Heaviside):
            self.expr = self.replace(Heaviside, UnitStep).expr
        if self.has(DiracDelta):
            self.expr = self.replace(DiracDelta, UnitImpulse).expr

    def first_index(self, nvals=None):

        if nvals is None:
            nvals = (-10, 10)                
        if isinstance(nvals, tuple):
            nvals = range(*nvals)

        # Desire SymPy equivalent to argmax
        n2 = nvals[-1]
        for n in reversed(nvals):
            if self(n) != 0:
                n2 = n
        return n2
        
    def last_index(self, nvals=None):

        if nvals is None:
            nvals = (-10, 10)        
        if isinstance(nvals, tuple):
            nvals = range(*nvals)

        # Desire SymPy equivalent to argmin
        n1 = nvals[0]
        for n in nvals:
            if self(n) != 0:
                n1 = n
        return n1

    def seq(self, nvals=None, evaluate=False):

        n1 = self.first_index(nvals)
        n2 = self.last_index(nvals)        

        # Perhaps if find self(n2 + 1) != 0 or self(n1 - 1) != 0 then
        # likely to have infinite extent sequence.  Maybe this could
        # be shown using ellipsis when the sequence is printed?
        
        nvals = arange(n1, n2 + 1)
        v = self(nvals)
        
        return Sequence(v, nvals, evaluate, self.var)
       



"""This module provides the TimeDomainExpression class to represent
time domain expressions.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import TimeDomain
from .expr import Expr
from .functions import exp
from .sym import fsym, ssym, tsym, j, oo, tausym
from .laplace import laplace_transform
from .fourier import fourier_transform
from sympy import Heaviside, limit, Integral, Expr as symExpr


__all__ = ('texpr', )

class TimeDomainExpression(TimeDomain, Expr):
    """t-domain expression or symbol."""

    var = tsym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)        
        assumptions['real'] = True
        super(TimeDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr        
        if check and expr.find(ssym) != set() and not expr.has(Integral):
            raise ValueError(
                't-domain expression %s cannot depend on s' % expr)
        if check and expr.find(fsym) != set() and not expr.has(Integral):
            raise ValueError(
                't-domain expression %s cannot depend on f' % expr)

    def _mul_compatible_domains(self, x):

        if self.domain == x.domain:
            return True        

        return x.is_constant_domain

    def _div_compatible_domains(self, x):

        if self.domain == x.domain:
            return True
        
        return x.is_constant_domain

    def as_expr(self):
        return TimeDomainExpression(self)

    def infer_assumptions(self):

        self.assumptions.infer_from_expr(self)

    def LT(self, evaluate=True, **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit.

        This is an alias for laplace."""

        return self.laplace(evaluate, **assumptions)
            
    def laplace(self, evaluate=True, **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit."""

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = laplace_transform(self.expr, self.var, ssym, evaluate=evaluate)
        return self.change(result, domain='laplace', **assumptions)

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        from .phasor import PhasorTimeDomainExpression
        
        return PhasorTimeDomainExpression.from_time(self, **assumptions)

    def FT(self, evaluate=True, **assumptions):
        """Attempt Fourier transform.  This is an alias for fourier.

        X(f) = \int_{-\infty}^{\infty} x(t) exp(-j 2\pi f t) dt."""

        return self.fourier(evaluate, **assumptions)        

    def fourier(self, evaluate=True, **assumptions):
        """Attempt Fourier transform."""

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = fourier_transform(self.expr, self.var, fsym, evaluate=evaluate)
        return self.change(result, domain='fourier', **assumptions)

    def angular_fourier(self, evaluate=True, **assumptions):
        """Attempt angular Fourier transform."""

        from .symbols import omega, pi, f

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = self.fourier(evaluate, **assumptions).subs(f, omega / (2 * pi))
        # Could optimise...
        return self.change(result, domain='angular fourier', **assumptions)

    def time(self, **assumptions):
        return self

    def plot(self, t=None, **kwargs):
        """Plot the time waveform.  If t is not specified, it defaults to the
        range (-0.2, 2).  t can be a vector of specified instants, a
        tuple specifing the range, or a constant specifying the
        maximum value with the minimum value set to 0.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        in addition to those supported by the matplotlib plot command.
        
        The plot axes are returned."""

        from .plot import plot_time
        return plot_time(self, t, **kwargs)

    def sample(self, t):

        """Return a discrete-time signal evaluated at time values specified by
        vector t. """

        return self.evaluate(t)

    def initial_value(self):
        """Determine value at t = 0. 
        See also pre_initial_value and post_initial_value"""

        return self.subs(0)

    def pre_initial_value(self):
        """Determine value at t = 0-.
        See also initial_value and post_initial_value"""

        return self.limit(self.var, 0, dir='-')

    def post_initial_value(self):
        """Determine value at t = 0+.
        See also pre_initial_value and initial_value"""

        return self.limit(self.var, 0, dir='+')    

    def final_value(self):
        """Determine value at t = oo."""

        return self.limit(self.var, oo)

    def remove_condition(self):
        """Remove the piecewise condition from the expression.
        See also force_causal."""

        if not self.is_conditional:
            return self
        expr = self.expr
        expr = expr.args[0].args[0]
        return self.__class__(expr)

    def force_causal(self):
        """Remove the piecewise condition from the expression
        and multiply by Heaviside function.  See also remove_condition."""

        if self.is_causal:
            return self
        
        expr = self.expr
        if self.is_conditional:
            expr = expr.args[0].args[0]            
        expr = expr * Heaviside(t)
        return self.__class__(expr)        

    def convolve(self, impulseresponse, commutate=False, **assumptions):
        """Convolve self with impulse response."""

        if not isinstance(impulseresponse, TimeDomainExpression):
            raise ValueError('Expecting TimeDomainExpression for impulse response')

        f1 = self.expr
        f2 = impulseresponse.expr
        if commutate:
            f1, f2 = f2, f1
        result = Integral(f1.subs(self.var, self.var - tausym) *
                          f2.subs(self.var, tausym),
                          (tausym, -oo, oo))
        return self.__class__(result, **assumptions)

    
class TimeDomainImpulseResponse(TimeDomainExpression):
    """Time-domain impulse response."""

    # TODO, check attributes.
    quantity = 'transfer'
    quantity_label = 'Impulse response'
    units = '1/s'
    is_transfer = True


def texpr(arg, **assumptions):
    """Create TimeDomainExpression object.  If `arg` is tsym return t"""

    if arg is tsym:
        return t
    return TimeDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses

classes = expressionclasses.register('time', TimeDomainExpression)
TimeDomainVoltage = classes['voltage']
TimeDomainCurrent = classes['current']

t = TimeDomainExpression('t')



"""This module performs nodal analysis.  It is primarily for showing
the equations rather than for evaluating them.

Copyright 2019-2020 Michael Hayes, UCECE

"""

from .circuitgraph import CircuitGraph
from .tmatrix import TimeDomainMatrix
from .utils import equation
from .systemequations import SystemEquations
import sympy as sym

__all__ = ('NodalAnalysis', )


class NodalAnalysis(object):
    """
    This is an experimental class for nodal analysis.  The API
    is likely to change.

    >>> from lcapy import Circuit, NodalAnalysis
    >>> cct = Circuit('''
    ... V1 1 0 {u(t)}; down
    ... R1 1 2; right=2
    ... L1 2 3; down=2
    ... W1 0 3; right
    ... W 1 5; up
    ... W 2 6; up
    ... C1 5 6; right=2
    ...''')    

    To perform nodal analysis in the Laplace domain:

    >>> na = NodalAnalysis(cct.laplace())
    
    To display the system of equations (in matrix form) that needs to
    be solved:
    
    >>> na.equations().pprint()
    
    To display the equations found by applying KCL at each node:
    
    >>> na.nodal_equations().pprint()

    """

    def __init__(self, cct, node_prefix=''):
        """`cct` is a netlist
        `node_prefix` can be used to avoid ambiguity between
        component voltages and node voltages."""

        self.cct = cct
        self.G = CircuitGraph(cct)

        self.kind = self.cct.kind
        if self.kind == 'super':
            self.kind = 'time'

        self.node_prefix = node_prefix
        
        self._unknowns = self._make_unknowns()
        
        self._y = matrix([val for key, val in self._unknowns.items() if key != '0'])
        
        self._equations = self._make_equations()

        
    def nodes(self):

        return self.G.nodes()

    def _make_unknowns(self):

        # Determine node voltage variables.
        unknowns = ExprDict()

        # cct.node_list is sorted alphabetically
        for node in self.cct.node_list:
            if node.startswith('*'):
                continue            
            if node == '0':
                unknowns[node] = 0
            else:
                unknowns[node] = Vname('V%s%s' % (self.node_prefix, node),
                                       self.kind)
        return unknowns

    def _make_equations(self):

        equations = {}
        for node in self.nodes():
            if node == '0':
                continue
            # Ignore dummy nodes
            if node.startswith('*'):
                continue

            voltage_sources = []
            for elt in self.G.connected(node):
                if elt.type == 'V':
                    voltage_sources.append(elt)

            if voltage_sources != []:
                elt = voltage_sources[0]
                n1 = self.G.node_map[elt.nodenames[0]]
                n2 = self.G.node_map[elt.nodenames[1]]

                V = elt.cpt.v_equation(0, self.kind)
                
                lhs, rhs = self._unknowns[n1], self._unknowns[n2] + V

            else:
                result = Itype(self.kind)(0)
                for elt in self.G.connected(node):
                    if len(elt.nodenames) < 2:
                        raise ValueError('Elt %s has too few nodes' % elt)
                    n1 = self.G.node_map[elt.nodenames[0]]
                    n2 = self.G.node_map[elt.nodenames[1]]
                    if node == n1:
                        pass
                    elif node == n2:
                        n1, n2 = n2, n1
                    else:
                        raise ValueError('Component %s does not have node %s' % (elt, node))
                    result += elt.cpt.i_equation(self._unknowns[n1] - self._unknowns[n2], self.kind)
                lhs, rhs = result, expr(0)

            equations[node] = (lhs, rhs)

        return equations        

    def equations_dict(self):
        """Return dictionary of equations keyed by node name."""

        equations_dict = ExprDict()
        for node, (lhs, rhs) in self._equations.items():
            equations_dict[node] = equation(lhs, rhs)

        return equations_dict
    
    def nodal_equations(self):
        """Return the equations found by applying KCL at each node.  This is a
        directory of equations keyed by the node name."""

        return self.equations_dict()

    def _analyse(self):

        if self.kind == 'time':
            raise ValueError('Cannot put time domain equations into matrix form')

        subsdict = {}
        for node, v in self._unknowns.items():
            if v == 0:
                continue
            subsdict[v.expr] = 'X_X' + node

        exprs = []
        for node, (lhs, rhs) in self._equations.items():
            lhs = lhs.subs(subsdict).expr.expand()
            rhs = rhs.subs(subsdict).expr.expand()            
            exprs.append(lhs - rhs)
            
        y = []
        for y1 in self._y:
            y.append(y1.subs(subsdict).expr);
        
        A, b = sym.linear_eq_to_matrix(exprs, *y)
        y = [y1.expr for y1 in self._y]
        return SystemEquations(A, b, y)

    @property
    def A(self):
        """Return A matrix where A y = b."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.A)

    @property
    def b(self):
        """Return b vector where A y = b."""        

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.b)        

    @property
    def y(self):
        """Return y vector where A y = b."""
        return self._y
    
    def matrix_equations(self, form='default', invert=False):
        """Return the equations in matrix form.

        Forms can be:
         A y = b
         b = A y
         Ainv b = y
         y = Ainv b

        If `invert` is True, evaluate the matrix inverse."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return self._sys.format(form, invert)

from .expr import ExprDict, expr
from .voltage import Vname
from .current import Itype
from .matrix import matrix



from __future__ import division
from .sym import symsimplify
from .functions import sqrt
from .sym import pi, omegasym, fsym
from .state import state
from .expr import Expr
import sympy as sym
import numpy as np

class NoiseExpression(Expr):
    """Frequency domain (one-sided) noise spectrum expression (amplitude
    spectral density).

    This characterises a zero-mean Gaussian noise process.

    When performing arithmetic on two noiseExpr expressions it is
    assumed that they are uncorrelated unless they have the same nid
    (noise indentifier).  If the nid is not specified, a new one is
    created.

    Uncorrelated noise expressions are added in quadrature (on a power
    basis).  Thus (NoiseExpression(3) + NoiseExpression(4)).expr = 5
    since 5 = sqrt(3**2 + 4**2)

    NoiseExpression(3) != NoiseExpression(3) since they are different
    noise realisations albeit with the same properties.  However,
    NoiseExpression(3).expr == NoiseExpression(3).expr.  Similarly,
    NoiseExpression(3, nid='n1') == NoiseExpression(3, nid='n1') since
    they have the same noise identifier and thus have the same
    realisation.

    Caution: The sum of two noise expressions generates a noise
    expression with a new nid.  This can lead to unexpected results
    since noise expressions with different nids are assumed to be
    uncorrelated.  For example, consider: 
    a = NoiseExpression(3); 
    b = NoiseExpression(4)
    a + b - b gives sqrt(41) and a + b - a gives sqrt(34).

    This case is correctly handled by the Super class since each noise
    component is stored and considered separately.

    (Voltage(a) + Voltage(b) - Voltage(b)).n gives 3 as expected.

    """
    is_real = True
    is_positive = True

    def _new_nid(self):
        state.context.nid += 1
        return 'n%d' % state.context.nid

    def __init__(self, val, **assumptions):
        if 'nid' not in assumptions or assumptions['nid'] is None:
            if val == 0:
                assumptions['nid'] = 'n0'
            else:
                assumptions['nid'] = self._new_nid()

        super(NoiseExpression, self).__init__(val, **assumptions)

    @property
    def nid(self):
        return self.assumptions['nid']

    def subs(self, *args, **kwargs):
        nid = self.nid
        result = super(NoiseExpression, self).subs(*args, **kwargs)
        result.assumptions['nid'] = nid
        return result

    def __compat__(self, x):

        if not isinstance(x, NoiseExpression):
            return x
        if self.var == x.var:
            return x
        return x(self.var)
    
    def __add__(self, x):
        """Add noise spectra (on power basis if uncorrelated)."""

        if not isinstance(x, NoiseExpression):
            raise ValueError('Cannot add %s and %s' % (self, x))
        
        if x == 0:
            return self.__class__(self, nid=self.nid)

        x = self.__compat__(x)
        
        if self.nid == x.nid:
            return self.__class__(self.expr + x.expr, nid=self.nid)
        
        value1 = self.expr
        value2 = x.expr
        value1sq = symsimplify(value1 * sym.conjugate(value1))
        value2sq = symsimplify(value2 * sym.conjugate(value2))                  
        result = symsimplify(sym.sqrt((value1sq + value2sq)))
        return self.__class__(result)

    def __radd__(self, x):
        raise ValueError('Cannot add %s and %s' % (self, x))        

    def __sub__(self, x):
        if not isinstance(x, NoiseExpression):
            raise ValueError('Cannot subtract %s and %s' % (self, x))

        if x == 0:
            return self.__class__(self, nid=self.nid)

        x = self.__compat__(x)        

        if self.nid == x.nid:
            return self.__class__(self.expr - x.expr, nid=self.nid)        
        return self + x

    def __rsub__(self, x):
        raise ValueError('Cannot subtract %s and %s' % (self, x))          

    def __mul__(self, x):
        if isinstance(x, NoiseExpression) and self.nid != x.nid:
            raise ValueError('Cannot multiply %s and %s' % (self, x))

        x = self.__compat__(x)
        return self.__class__(self.expr * x, nid=self.nid)

    def __rmul__(self, x):
        return self.__class__(self.expr * x, nid=self.nid)    

    def __div__(self, x):
        if isinstance(x, NoiseExpression) and self.nid != x.nid:
            raise ValueError('Cannot divide %s and %s' % (self, x))
        return self.__class__(self.expr / x, nid=self.nid)

    def __rdiv__(self, x):
        return self.__class__(x / self.expr, nid=self.nid)        

    def __eq__(self, x):
        try:
            if self.nid != x.nid:
                return False
        except:
            pass

        x = self.__compat__(x)        
        try:
            if self.expr == x.expr:
                return True
        except:
            pass
        return self.expr == x

    def __ne__(self, x):
        return not (self == x)

    def rms(self):
        """Calculate rms value."""

        P = sym.integrate(self.expr**2, (self.var, 0, sym.oo))
        if self.var == omegasym:
            P /= 2 * sym.pi
        rms = sym.sqrt(P)
        # TODO: Use rms class?
        cls = self._class_by_quantity(self.quantity, 'time')
        return cls(rms)

    def sample(self, t, seed=None):
        """Return a sample function (realisation) of the noise process
        evaluated at time values specified by vector t.   If t is an integer,
        this returns `t` samples."""

        if isinstance(t, int):
            t = np.arange(t)
        
        N = len(t)
        if N < 3:
            raise ValueError('Require at least 3 samples')
        if N & 1:
            raise ValueError('Require even number of samples')

        if seed is not None:
            np.random.seed(seed)
        
        td = np.diff(t)
        if not np.allclose(np.diff(td), 0):
            raise ValueError('Require uniform sampling')

        fs = 1 / td[0]
        vf = np.arange(N // 2 + 1) * fs / N
        Sn = self(f).evaluate(vf)
    
        x = np.random.randn(N)
        X = np.fft.rfft(x)
        
        Y = X * np.sqrt(Sn * fs / 2)
        y = np.fft.irfft(Y)
        return y

    def time(self):
        print('Warning: no time representation for noise expression, '
              'assumed zero: use rms()')
        return 0

    def fourier(self):
        print('Warning: no Fourier representation for noise expression, '
              'assumed zero: use asd() or psd()')
        return 0    

    def laplace(self):
        print('Warning: no Laplace representation for noise expression, '
              'assumed zero')
        return 0

    def asd(self):
        return self

    def psd(self):
        # FIXME, the units are wrong
        return self ** 2

    def autocorrelation(self):
        # Convert to two-sided spectrum
        S = self.subs(self.var, abs(self.var)) / sqrt(2)
        return S.inverse_fourier()

    def plot(self, fvector=None, **kwargs):
        """Plot frequency response at values specified by fvector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.
        """
        from .plot import plot_frequency, plot_angular_frequency

        if self.var == fsym:
            return plot_frequency(self, fvector, **kwargs)
        elif self.var == omegasym:
            return plot_angular_frequency(self, fvector, **kwargs)
        raise ValueError('Invalid var for NoiseExpression')

    
from .fexpr import f
from .omegaexpr import omega



"""
This module performs state-space analysis.

Copyright 2019-2020 Michael Hayes, UCECE

"""

from .mnacpts import I, V, C, L
from .expr import expr
from .matrix import Matrix
from .smatrix import LaplaceDomainMatrix
from .tmatrix import TimeDomainMatrix
from .voltage import voltage
from .current import current
from .sym import sympify, ssym
import sympy as sym

__all__ = ('StateSpace', )

# TODO
# 1. Use a better Matrix class that preserves the class of each
# element, where possible.
# 2. Handle namespaces

# Independent sources
# V1 1 0
# V1 1 0 {10 * u(t)}
# V1 1 0 ac 10
#
# In each case, we use v1(t) as the independent source when determining
# the A, B, C, D matrices.  We can then substitute the known value at
# the end.


def _hack_vars(exprs):
    """Substitute i_Canon1(t) with i_C(t) etc. provided
    there is no i_Canon2(t)."""

    for m, expr1 in enumerate(exprs):
        for c in ('i_V', 'i_C', 'i_L', 'v_C'):
            sym1 = sympify(c + 'anon1(t)')
            sym2 = sympify(c + 'anon2(t)')            
            if expr1.has(sym1) and not expr1.has(sym2):
                expr1 = expr1.subs(sym1, sympify(c + '(t)'))
                exprs[m] = expr1

class StateSpace(object):
    """This converts a circuit to state-space representation.

    It can take a long time for a symbolic circuit with many reactive
    components."""

    def __init__(self, cct, node_voltages=True, branch_currents=False):

        if not node_voltages and not branch_currents:
            raise ValueError('No outputs')
        
        inductors = []
        capacitors = []
        independent_sources = []

        # Determine state variables (current through inductors and
        # voltage across acapacitors) and replace inductors with
        # current sources and capacitors with voltage sources.
        sscct = cct._new()
        cpt_map = {}

        for key, elt in cct.elements.items():
            ssnet = elt._ss_model()
            sselt = sscct._add(ssnet)
            name = elt.name
            cpt_map[name] = sselt.name
            
            if elt.is_inductor:
                if sselt.name in cct.elements:
                    raise ValueError('Name conflict %s, either rename the component or improve the code!' % sselt.name)

                inductors.append(elt)
            elif elt.is_capacitor:
                if sselt.name in cct.elements:
                    raise ValueError('Name conflict %s, either rename the component or improve the code!' % sselt.name)
                
                capacitors.append(elt)
            elif isinstance(elt, (I, V)):
                independent_sources.append(elt)                
                
        self.cct = cct
        self.sscct = sscct
        # sscct can be analysed in the time domain since it has no
        # reactive components.  However, for large circuits
        # this can take a long time due to inversion of the MNA matrix.
        
        dotx_exprs = []
        statevars = []
        statenames = []
        initialvalues = []
        for elt in inductors + capacitors:
            name = cpt_map[elt.name]

            if isinstance(elt, L):
                # Inductors  v = L di/dt  so need v across the L
                expr = sscct[name].v / elt.cpt.L
                var = -sscct[name].isc
                x0 = elt.cpt.i0
            else:
                # Capacitors  i = C dv/dt  so need i through the C
                # The current is negated since it is from a source V_Cx
                expr = -sscct[name].i / elt.cpt.C
                var = sscct[name].voc
                x0 = elt.cpt.v0

            dotx_exprs.append(expr)
            statevars.append(var)
            statenames.append(name)
            initialvalues.append(x0)

        statesyms = sympify(statenames)

        # Determine independent sources.
        sources = []
        sourcevars = []
        sourcenames = []
        for elt in independent_sources:
            name = cpt_map[elt.name]
            
            if isinstance(elt, V):
                expr = elt.cpt.voc
                var = sscct[name].voc                
            else:
                expr = elt.cpt.isc
                var = sscct[name].isc

            sources.append(expr)
            sourcevars.append(var)
            sourcenames.append(name)

        sourcesyms = sympify(sourcenames)            

        # linear_eq_to_matrix expects only Symbols and not AppliedUndefs,
        # so substitute.
        subsdict = {}
        for var, sym1 in zip(statevars, statesyms):
            subsdict[var.expr] = sym1
        for var, sym1 in zip(sourcevars, sourcesyms):
            subsdict[var.expr] = sym1       

        for m, expr in enumerate(dotx_exprs):
            dotx_exprs[m] = expr.subs(subsdict).expr.expand()

        A, b = sym.linear_eq_to_matrix(dotx_exprs, *statesyms)
        if sourcesyms != []:
            B, b = sym.linear_eq_to_matrix(dotx_exprs, *sourcesyms)
        else:
            B = []

        # Determine output variables.
        yexprs = []
        y = []

        if node_voltages:
            for node in cct.node_list:
                if node == '0':
                    continue
                yexprs.append(self.sscct[node].v.subs(subsdict).expand().expr)
                # Note, this can introduce a name conflict
                y.append(voltage('v_%s(t)' % node))

        if branch_currents:
            for name in cct.branch_list:
                # Perhaps ignore L since the current through it is a
                # state variable?
                name2 = cpt_map[name]                    
                yexprs.append(self.sscct[name2].i.subs(subsdict).expand().expr)
                y.append(current('i_%s(t)' % name))                    

        Cmat, b = sym.linear_eq_to_matrix(yexprs, *statesyms)
        if sourcesyms != []:        
            D, b = sym.linear_eq_to_matrix(yexprs, *sourcesyms)
        else:
            D = []

        # Rewrite vCanon1(t) as vC(t) etc if appropriate.
        _hack_vars(statevars)
        _hack_vars(sources)
        
        # Note, Matrix strips the class from each element...
        self.x = TimeDomainMatrix(statevars)

        self.x0 = Matrix(initialvalues)
        
        self.dotx = TimeDomainMatrix([sym.Derivative(x1, t) for x1 in self.x])

        self.u = TimeDomainMatrix(sources)

        self.A = Matrix(A)
        self.B = Matrix(B)        
            
        # Perhaps could use v_R1(t) etc. as the output voltages?
        self.y = TimeDomainMatrix(y)

        self.C = Matrix(Cmat)
        self.D = Matrix(D)

    def state_equations(self):
        """System of first-order differential state equations:

        dotx = A x + B u

        where x is the state vector and u is the input vector.
        """
        
        return expr(sym.Eq(self.dotx, sym.MatAdd(sym.MatMul(self.A, self.x),
                                                 sym.MatMul(self.B, self.u)),
                           evaluate=False))

    def output_equations(self):
        """System of output equations:

        y = C x + D u

        where y is the output vector, x is the state vector and u is
        the input vector.

        """
        
        return expr(sym.Eq(self.y, sym.MatAdd(sym.MatMul(self.C, self.x),
                                              sym.MatMul(self.D, self.u)),
                           evaluate=False))

    @property
    def Phi(self):
        """s-domain state transition matrix."""

        M = LaplaceDomainMatrix(sym.eye(len(self.x)) * ssym - self.A)
        return LaplaceDomainMatrix(M.inv().canonical())

    @property
    def phi(self):
        """State transition matrix."""        
        return TimeDomainMatrix(self.Phi.inverse_laplace(causal=True))
        
    @property
    def U(self):
        """Laplace transform of input vector."""
        return LaplaceDomainMatrix(self.u.laplace())

    @property
    def X(self):
        """Laplace transform of state-variable vector."""        
        return LaplaceDomainMatrix(self.x.laplace())

    @property
    def Y(self):
        """Laplace transform of output vector."""        
        return LaplaceDomainMatrix(self.y.laplace())

    @property
    def H(self):
        """X(s) / U(s)"""

        return LaplaceDomainMatrix(self.Phi * self.B).canonical()

    @property
    def h(self):
        return TimeDomainMatrix(self.H.inverse_laplace(causal=True))

    @property
    def G(self):
        """System transfer functions."""

        return LaplaceDomainMatrix(self.C * self.H + self.D).canonical()

    @property
    def g(self):
        """System impulse responses."""        
        return TimeDomainMatrix(self.G.inverse_laplace(causal=True))
    
    def characteristic_polynomial(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(s) = |s * I - A|`"""

        M = Matrix(sym.eye(len(self.x)) * ssym - self.A)        
        return LaplaceDomainExpression(M.det()).simplify()

    @property
    def P(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(s) = |s * I - A|`"""        

        return self.characteristic_polynomial().canonical()

    @property        
    def eigenvalues_dict(self):
        """Dictionary of eigenvalues, the roots of the characteristic
        polynomial (equivalent to the poles of Phi(s)).  The
        dictionary values are the multiplicity of the eigenvalues.

        For a list of eigenvalues use eigenvalues."""        

        return self.characteristic_polynomial().roots()
        
    @property        
    def eigenvalues(self):
        """List of eigenvalues, the roots of the characteristic polynomial
        (equivalent to the poles of Phi(s))."""
        
        roots = self.eigenvalues_dict
        e = []

        # Replicate duplicated eigenvalues and return as a list.
        for v, n in roots.items():
            for m in range(n):
                e.append(v)
        return ExprList(e)

    @property    
    def Lambda(self):
        """Diagonal matrix of eigenvalues."""

        # Perhaps faster to use diagonalize
        # E, L = self.A.diagonalize()
        # return L
        
        e = self.eigenvalues
        return LaplaceDomainMatrix(sym.diag(*e))

    @property        
    def eigenvectors(self):
        """List of tuples (eigenvalue, multiplicity of eigenvalue,
        basis of the eigenspace) of A."""
        
        return self.A.eigenvects()
    
    @property    
    def M(self):
        """Modal matrix (eigenvectors of A)."""

        E, L = self.A.diagonalize()
        
        return LaplaceDomainMatrix(E)
    
    
from .symbols import t, s
from .expr import ExprList
from .sexpr import LaplaceDomainExpression



from lcapy import *
from lcapy.superpositionvoltage import SuperpositionVoltage
from lcapy.superpositioncurrent import SuperpositionCurrent
import unittest
import sympy as sym

def div(a, b):
   return a / b

def mul(a, b):
   return a * b

class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_voltage_properties(self):
        self.assertEqual(SuperpositionVoltage(3).is_dc, True, "Voltage(3).is_dc")
        self.assertEqual(SuperpositionVoltage(phasorvoltage(3)).is_ac, True, "Voltage(Vphasor(3)).is_ac")
        self.assertEqual(SuperpositionVoltage(voltage(2), phasorvoltage(3)).is_ac, False,
                          "Voltage(Vconst(2), Vphasor(3)).is_ac")
        self.assertEqual(SuperpositionVoltage(voltage(2), phasorvoltage(3)).is_ac, False,
                          "Voltage(Vconst(2), Vphasor(3)).is_dc")

    def test_voltage_add_sub_dc(self):
        self.assertEqual2(SuperpositionVoltage(3).dc, 3, "Voltage(3).dc")
        self.assertEqual2(SuperpositionVoltage(2, 3).dc, 5, "Voltage(2, 3).dc")
        self.assertEqual2(SuperpositionVoltage(2, 3).ac, {}, "Voltage(2, 3).ac")
        self.assertEqual2(-SuperpositionVoltage(2).dc, -2, "-Voltage(2).dc")
        self.assertEqual2(SuperpositionVoltage(2) + SuperpositionVoltage(3), SuperpositionVoltage(5),
                          "Voltage(2) + Voltage(3)")
        self.assertEqual2(SuperpositionVoltage(2) - SuperpositionVoltage(3), SuperpositionVoltage(-1),
                          "Voltage(2) - Voltage(3)")

    def test_current_add_sub_dc(self):
        self.assertEqual2(SuperpositionCurrent(3).dc, 3, "Current(3).dc")
        self.assertEqual2(SuperpositionCurrent(2, 3).dc, 5, "Current(2, 3).dc")
        self.assertEqual2(SuperpositionCurrent(2, 3).ac, {}, "Current(2, 3).ac")
        self.assertEqual2(-SuperpositionCurrent(2).dc, -2, "-Current(2).dc")
        self.assertEqual2(SuperpositionCurrent(2) + SuperpositionCurrent(3), SuperpositionCurrent(5),
                          "Current(2) + Current(3)")
        self.assertEqual2(SuperpositionCurrent(2) - SuperpositionCurrent(3), SuperpositionCurrent(-1),
                          "Current(2) - Current(3)")
        
    #def test_Voltage_noise(self):
        # self.assertEqual((AngularFourierNoiseDomainVoltage(3) + AngularFourierNoiseDomainVoltage(4)).expr, AngularFourierNoiseDomainVoltage(5).expr, "Vnoisy(3) + Vnoisy(4)")
        # self.assertEqual((SuperpositionVoltage(AngularFourierNoiseDomainVoltage(3)) + SuperpositionVoltage(AngularFourierNoiseDomainVoltage(4))).n.expr,
        #                   SuperpositionVoltage(AngularFourierNoiseDomainVoltage(5)).n.expr,
        #                   "Voltage(Vnoisy(3)) + Voltage(Vnoisy(4))")

        
    def test_voltage_has(self):

        a = SuperpositionVoltage('3 * exp(-t) * t * a')
        self.assertEqual(a.has(3), True, "has(3)")
        self.assertEqual(a.has(4), False, "has(4)")
        self.assertEqual(a.has(t), True, "has(t)")
        self.assertEqual(a.has_symbol(t), True, "has_symbol(t)")
        self.assertEqual(a.has_symbol('a'), True, "has_symbol(a)")
        self.assertEqual(a.has_symbol('b'), False, "has_symbol(b)")

    def test_voltage_transform(self):

        V1 = SuperpositionVoltage('3 * exp(-2 * t)')
        self.assertEqual(V1.transform(s), 3 / (s + 2), 'transform(s)')        

        V2 = SuperpositionVoltage('3 * exp(-2 * t) * u(t)')
        self.assertEqual(V2.transform(s), 3 / (s + 2), 'transform(s)')        
        self.assertEqual(simplify(V2.transform(f) - 3 / (j * 2 * pi * f + 2)), 0, 'transform(f)')                
        
    def test_voltage_subs(self):

        a = SuperpositionVoltage('V1')
        b = a.subs('V1', 1)
        c = SuperpositionVoltage(1)
        self.assertEqual(b, c, "Voltage.subs")
        
    def test_voltage_decompose(self):

        V1 = SuperpositionVoltage('1 + 3 * u(t) + cos(2 * pi * 3 * t)')
        self.assertEqual(V1.dc, 1, '.dc')
        self.assertEqual(V1.transient, expr('3 * u(t)'), '.transient')

    def test_voltage_oneport(self):

        V1 = V(3)

        self.assertEqual(V1.V.oneport().V, V1.V, 'oneport')

    def test_current_oneport(self):

        I1 = I(3)

        self.assertEqual(I1.I.oneport().I, I1.I, 'oneport')                
        
    def test_Vname(self):

        self.assertEqual(Vname('V', 't'), voltage('v(t)'), 'v(t)')
        self.assertEqual(Vname('V', 's'), voltage('V(s)'), 'V(s)')
        # TODO: remove cache requirement
        self.assertEqual(Vname('V', 'dc', cache=True), voltage('V'), 'V')        

    def test_Iname(self):
        
        self.assertEqual(Iname('I', 't'), current('i(t)'), 'i(t)')
        self.assertEqual(Iname('I', 's'), current('I(s)'), 'I(s)')
        self.assertEqual(Iname('I', 'dc', cache=True), current('I'), 'I')
        
    def test_voltage_phasor(self):

        V = SuperpositionVoltage(3 * sin(7 * t) + 2 * cos(14 * t)) 
        self.assertEqual(V[7].magnitude, expr(3), 'magnitude')
        self.assertEqual(V[14].omega, 14, 'omega')
        
    def test_super_printing(self):

        self.assertEqual(SuperpositionVoltage(3).latex(), '3', 'SuperpositionVoltage(3)')
        self.assertEqual(SuperpositionVoltage(3 * s).latex(), '3 s', 'SuperpositionVoltage(3 * s)')
        self.assertEqual(SuperpositionVoltage(2 * sin(omega * t)).latex(),
                         '2 \\sin{\\left(\\omega t \\right)}',
                         'SuperpositionVoltage(2 * sin(omega * t))')
        self.assertEqual(SuperpositionVoltage(3 + 2 * sin(omega * t)).latex(),
                         '2 \\sin{\\left(\\omega t \\right)} + 3',
                         'SuperpositionVoltage(3 + 2 * sin(omega * t))')        

    def test_super_op_const(self):

        V = SuperpositionVoltage(10)
        Z = impedance(2)
        Y = admittance(1 / 2)
        I = SuperpositionCurrent(5)
       
        self.assertEqual(V._div(Z), I, 'V / R')
        self.assertEqual(I._div(Y), V, 'I / G')
        self.assertEqual(V._mul(Y), I, 'V / G')
        self.assertEqual(I._mul(Z), V, 'I * R')

    def test_super_op_laplace(self):

        V = SuperpositionVoltage(10 / s)
        Z = impedance(2 * s)
        Y = admittance(1 / (2 * s))
        I = SuperpositionCurrent(5 / s**2)
       
        self.assertEqual(V._div(Z), I, 'V / R')
        self.assertEqual(I._div(Y), V, 'I / G')
        self.assertEqual(V._mul(Y), I, 'V / G')
        self.assertEqual(I._mul(Z), V, 'I * R')

    def test_super_op_phasor(self):

        V = SuperpositionVoltage((10 * sin(omega * t)).as_phasor())
        # TODO, try with jw
        Z = impedance(2 * s)
        Y = admittance(1 / (2 * s))
        I = SuperpositionCurrent((-5 / omega * cos(omega * t)).as_phasor())
       
        self.assertEqual(V._div(Z), I, 'V / R')
        self.assertEqual(I._div(Y), V, 'I / G')
        self.assertEqual(V._mul(Y), I, 'V / G')
        self.assertEqual(I._mul(Z), V, 'I * R')                

    def test_super_time_div_laplace(self):

        self.assertRaises(ValueError, div, voltage('v(t)'), impedance(3 * s))

    def test_super_time_mul_laplace(self):

        self.assertRaises(ValueError, div, current('i(t)'), admittance(3 * s))                

    def test_super_mul(self):

        self.assertEqual(SuperpositionCurrent(5) * impedance(2), SuperpositionVoltage(10), 'I * R')
        self.assertEqual(SuperpositionCurrent(5) * impedance(0 * s + 2), SuperpositionVoltage(10), 'I * R')

    def test_super_reverse_mul(self):

        self.assertEqual(impedance(2) * SuperpositionCurrent(5), SuperpositionVoltage(10), 'I * R')

    def test_super_sympy_float(self):

        V = SuperpositionVoltage()
        V.add(sym.Float(8))
       
        self.assertEqual(isinstance(V.dc.expr, sym.Rational), True, 'Float -> Rational')



"""This module provides functions.

Copyright 2014--2020 Michael Hayes, UCECE
"""


import sympy as sym

class Function(object):

    def __init__(self, arg):
        self.expr = arg
    
    def __call__(self, *args):

        cls = args[0].__class__

        # Unwrap expressions
        tweak_args = list(args)
        for m, arg in enumerate(args):
            if isinstance(arg, (Expr, Function)):
                tweak_args[m] = arg.expr

        result = self.expr(*tweak_args)

        if isinstance(args[0], Expr):

            result = cls(result)
            if args[0].is_phase and self.expr in (sym.sin, sym.cos, sym.tan, sym.exp,
                                                  sym.sinh, sym.cosh, sym.tanh):
                result.part = ''

        if False:
            for m, arg in enumerate(args[1:]):
                if isinstance(arg, (Expr, Function)):
                    # Need to avoid substituting constants
                    result = result.subs(tweak_args[m], arg)

        return expr(result)

    def pdb(self):
        import pdb; pdb.set_trace()
        return self
    
    
class Log10(Function):

    # TODO, figure out how to print as log10(x) rather than
    # the expansion log(x) / log(10).  This will require
    # deferment of the expansion.
    
    def __call__(self, arg):
        return super(Log10, self).__call__(arg, 10)

   
conjugate = Function(sym.conjugate)

sqrt = Function(sym.sqrt)

exp = Function(sym.exp)

log = Function(sym.log)

log10 = Log10(sym.log)

sin = Function(sym.sin)

cos = Function(sym.cos)

tan = Function(sym.tan)

cot = Function(sym.cot)

asin = Function(sym.asin)

acos = Function(sym.acos)

atan = Function(sym.atan)

atan2 = Function(sym.atan2)

acot = Function(sym.acot)

sinh = Function(sym.sinh)

cosh = Function(sym.cosh)

tanh = Function(sym.tanh)

asinh = Function(sym.asinh)

acosh = Function(sym.acosh)

atanh = Function(sym.atanh)

gcd = Function(sym.gcd)

sign = Function(sym.sign)

u = H = heaviside = Heaviside = Function(sym.Heaviside)

delta = DiracDelta = Function(sym.DiracDelta)


# def delta(expr1):
#    
#     if not hasattr(expr1, 'expr'):
#         expr1 = expr(expr1)
#     if expr1.is_discrete_time_domain or expr1.is_discrete_fourier_domain:
#         return UnitImpulse(expr1)        
#    
#     return DiracDelta(expr1)


def _ex(expr):
    if hasattr(expr, 'expr'):
        return expr.expr
    return expr


class Eq(sym.Eq):
    def __new__(cls, lhs, rhs=None, **options):
        return expr(super(Eq, cls).__new__(cls, _ex(lhs), _ex(rhs), **options))


class Add(sym.Add):
    def __new__(cls, op1, op2, **options):
        return expr(super(Add, cls).__new__(cls, _ex(op1), _ex(op2), **options))

    
class Mul(sym.Mul):
    def __new__(cls, op1, op2, **options):
        return expr(super(Mul, cls).__new__(cls, _ex(op1), _ex(op2), **options))

    
class MatAdd(sym.MatAdd):
    def __new__(cls, op1, op2, **options):
        return expr(super(MatAdd, cls).__new__(cls, _ex(op1), _ex(op2), **options))

    
class MatMul(sym.MatMul):
    def __new__(cls, op1, op2, **options):
        return expr(super(MatMul, cls).__new__(cls, _ex(op1), _ex(op2), **options))


from .extrafunctions import UnitImpulse, UnitStep
from .extrafunctions import sinc as sinc1
from .extrafunctions import rect as rect1


ui = unitimpulse = Function(UnitImpulse)

us = unitstep = Function(UnitStep)

rect = Function(rect1)

sinc = Function(sinc1)
 

from .expr import Expr, expr
